unit chatapp_class;

{
Программа "Видеочат"
Copyright (c) 2020 Любезный В. В.
https://www.lubezniy.ru/soft/videochat

Модуль с описанием класса приложения чата
}

{$mode objfpc}{$H+}

interface

uses Classes, webserverconfig_class, db_users, db_sessions, db_messages;

type

  TChatApp = class
    // Объекты конфигурации
    WebServerConfig: TWebserverConfig;
    // Объект конфигурации Web-сервера
    UserDB: TUserDB; // Информация о пользователях

    // Объекты частей приложения
    SessionsDB: TSessionsDB; // База данных сессий
    MessagesDB: TMessagesDB; // База данных сообщений чата


    // Методы создания объектов, запуска и остановки приложения
    constructor Create;
    destructor Destroy;
    procedure Start;

    // Метод, отдающий индекс пользователя в БД по уникальному идентификатору сессии (-1, если сессии с таким идентификатором в базе нет)
    function getUserIndexBySession(Session: string): integer;

    // Метод получения истории чата и оформления в виде JSON-кода в TStringList
    function getChatHistory(currentUserUID, contactUID: string;
      currentTimestamp: int64): TStringList;

    // Метод проверки наличия и получения новых сообщений в виде JSON-кода в TStringList
    function getNewMessages(currentUserUID: string): TStringList;

    // Метод получения текущего состояния для строки в приложении
    function getState: string;
  end;

implementation

uses SysUtils, fpjson;

constructor TChatApp.Create;
begin
  inherited Create;
end;

destructor TChatApp.Destroy;
begin
  inherited Destroy;
end;

procedure TChatApp.Start;
var
  s: string;
begin
  // Читаем конфигурацию и данные пользователей.
  WebServerConfig.Read;
  UserDB.Read;

  // При отсутствии каталога БД создаём его
  s := ExtractFilePath(ParamStr(0)) + 'database';
  if not DirectoryExists(s) then
    MkDir(s);

  // Работаем по объектам БД
  SessionsDB.readFromFile;
  MessagesDB.readFromFile;
end;

function TChatApp.getUserIndexBySession(Session: string): integer;
var
  UserUID: string;
begin
  // Пока мы ничего не нашли, возвращать будем -1.
  Result := -1;
  if Session = '' then
  begin
    Exit;
  end;
  // Проверяем сессию, получаем уникальный идентификатор пользователя.
  UserUID := SessionsDB.getUserUID(Session);
  if UserUID <> '' then
  begin
    // Нашли - получаем индекс пользователя в базе
    Result := UserDB.IndexByUID(UserUID);
  end;

  if Result <> -1 then
  begin
    // Если нашли, устанавливаем в базе дату и время последнего логина.
    UserDB.Recs[Result].lastLogin := Trunc((Now - EncodeDate(1970, 1, 1)) * 24 * 60 * 60);
  end;
end;

function TChatApp.getChatHistory(currentUserUID, contactUID: string;
  currentTimestamp: int64): TStringList;
var
  i: integer;
  History: TMessageRecArray;
  s: string;
  b: boolean;
  t: int64;
  jo: TJSONObject;
begin
  // Создаём объект типа TStringList для записи результата и на всякий случай принудительно очищаем его.
  Result := TStringList.Create;
  Result.Clear;
  // Добавляем первую строку JSON - code 0 и начало массива
  Result.Add('{"code": 0, "data": [');
  // Получаем историю сообщений из базы в массив History
  MessagesDB.getHistory(currentUserUID, contactUID, History);
  if Length(History) > 0 then
  begin
    // История не пустая - начинаем выдачу в обратном порядке.
    // С помощью переменной t определяем время последнего сообщения в истории.
    t := 0;
    b := False;
    for i := Length(History) - 1 downto 0 do
    begin
      if History[i].when > t then
      begin
        t := History[i].when;
      end;
      // В переменной s формируем строку для TStringList с записью истории сообщений.
      if b then
      begin
        // Для второго и последующих элементов выдачи начинаем строку с запятой.
        s := ',';
      end
      else
      begin
        // Первому элементу запятая не нужна - строка будет просто пустой.
        s := '';
        b := True;
      end;
      // Создаём JSON-объект с записью истории и заполняем его значениями полей этой записи.
      jo := TJSONObject.Create;
      jo.Add('when', History[i].when);
      jo.Add('text', History[i].Text);
      jo.Add('msgType', History[i].msgType);
      jo.Add('sender', History[i].Sender);
      jo.add('recipient', History[i].recipient);
      // Добавляем строковое представление JSON-объекта к переменной s
      s := s + jo.AsJSON;
      // Освобождаем память от JSON-объекта
      FreeAndNil(jo);
      // Добавляем строку в наш TStringList и переходим к следующей записи истории, если таковая будет.
      Result.Add(s);
    end;
  end
  else
  begin
    // История пуста - временем последнего сообщения ставим текущее.
    t := currentTimestamp;
  end;
  // Добавляем в наш результат последнюю строку - завершение массива и элемент с макс. временем.
  Result.Add('], "lastTimestamp": ' + IntToStr(t) + '}');
end;

function TChatApp.getNewMessages(currentUserUID: string): TStringList;
var
  i: integer;
  Messages: TMessageRecArray;
  s: string;
  b: boolean;
  jo: TJSONObject;
begin
  // Создаём объект типа TStringList для записи результата и на всякий случай принудительно очищаем его.
  Result := TStringList.Create;
  Result.Clear;
  // Добавляем первую строку JSON - code 0 и начало массива
  Result.Add('{"code": 0, "data": [');
  // Получаем новые сообщения в массив Messages
  MessagesDB.getNew(currentUserUID, Messages);
  if Length(Messages) > 0 then
  begin
    // Массив новых сообщений не пуст - отдаём его в обратном порядке.
    b := False;
    for i := Length(Messages) - 1 downto 0 do
    begin
      // В переменной s формируем строку для TStringList с записью нового сообщения.
      if b then
      begin
        // Для второго и последующих элементов выдачи начинаем строку с запятой.
        s := ',';
      end
      else
      begin
        // Первому элементу запятая не нужна - строка будет просто пустой.
        s := '';
        b := True;
      end;
      // Создаём JSON-объект с записью и заполняем его значениями полей этой записи.
      jo := TJSONObject.Create;
      jo.Add('when', Messages[i].when);
      jo.Add('text', Messages[i].Text);
      jo.Add('msgType', Messages[i].msgType);
      jo.Add('sender', Messages[i].Sender);
      jo.add('recipient', Messages[i].recipient);
      // Добавляем строковое представление JSON-объекта к переменной s
      s := s + jo.AsJSON;
      // Освобождаем память от JSON-объекта
      FreeAndNil(jo);
      // Добавляем строку в результат и переходим к следующей записи
      Result.Add(s);
    end;
  end;
  // Добавляем строку для завершения JSON-массива.
  Result.Add(']}');
end;

function TChatApp.getState: string;
begin
  Result := 'Всего пользователей: ' +
    IntToStr(Length(UserDB.Recs)) + ', онлайн ' + IntToStr(UserDB.onlineCount);
end;

end.
