unit db_messages;

{
Программа "Видеочат"
Copyright (c) 2020 Любезный В. В.
https://www.lubezniy.ru/soft/videochat

Модуль описания сущностей и класса базы данных сообщений чата.
}

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils;

type

  // Тип записи сообщения
  TMessageRec = record
    when: int64; // UNIX-timestamp получения сообщения
    Text: string; // текст сообщения
    msgType: byte;
    // тип сообщения (0 - текст, 1 - звонок, 2 - приём видеозвонка, 3 - отклонение звонка, 4 - offer, 5 - answer, 6 - ICE-кандидат, 7 - сообщение о завершении звонка, 8 - отказ пользователя от исходящего звонка до его начала, 9 - перезапуск ICE)
    Sender: string[32];
    // Уникальный идентификатор пользователя - отправителя сообщения из БД пользователей
    recipient: string[32];
    // Уникальный идентификатор пользователя - получателя сообщения из БД пользователей
    uid: string[32]; // Уникальный идентификатор сообщения
    // Флаг о том, что сообщение получено
    gotByRecipient: boolean;
  end;

  // Тип массива записей сообщений
  TMessageRecArray = array of TMessageRec;

  // Тип объекта базы данных сообщений
  TMessagesDB = object

    // Массив сообщений
    Recs: TMessageRecArray; // записи

    // Метод инициализации БД
    procedure Init;

    // Чтение БД сообщений из файла (временно не задействовано)
    procedure readFromFile;

    {
     Добавление сообщения.
     Параметры:

      when - UNIX-timestamp даты-времени сообщения (кол-во секунд, прошедших с 0:00 1 января 1970 года до заданного времени);
      msgType - тип сообщения (см. описание tMessageRec)
      senderUID - уникальный идентификатор пользователя - отправителя сообщения (из БД пользователей)
      recipientUID - уникальный идентификатор пользователя - получателя сообщения (из БД пользователей)

     В переменную recUID метод запишет уникальный идентификатор сообщения в массиве Recs. В переменную recIndex - индекс сообщения в этом же массиве).
    }
    procedure addMessage(when: int64; senderUID, recipientUID: string;
      msgType: byte; msgText: string; var recIndex: integer; var recUID: string);

    {
     Получение истории сообщений между абонентом и его собеседником.
     Параметры:
      abonentUID - уникальный идентификатор пользователя-абонента
      contactUID - уникальный идентификатор пользователя-собеседника

     Результат работы метод запишет в переменную History
    }
    procedure getHistory(abonentUID, contactUID: string; var History: TMessageRecArray);

    {
     Получение новых сообщений для абонента.
     Параметры:
      abonentUID - уникальный идентификатор пользователя-абонента

     Результат работы метод запишет в переменную Messages
    }
    procedure getNew(abonentUID: string; var Messages: TMessageRecArray);
    // получение новых сообщений

    {
     Получение индекса записи сообщения в базе по его уникальному идентификатору. При отсутствии сообщения с таким идентификатором возвращается -1.
    }
    function indexByUID(UID: string): integer;

    {
     Запись БД сообщений в файл. Временно не задействован.
    }
    procedure saveToFile;
  end;

implementation

uses stringfuncs;

procedure TMessagesDB.Init;
begin
  SetLength(Recs, 0);
end;

procedure TMessagesDB.getHistory(AbonentUID, ContactUID: string;
  var History: TMessageRecArray);
var
  i, j: longint;
begin
  // Очищаем массив результата.
  SetLength(History, 0);
  if Length(Recs) > 0 then
  begin
    // База сообщений не пуста - можно перебирать.
    // Переменная j - счётчик кол-ва сообщений в истории
    j := 0;
    // Отсчёт записей ведём в обратном порядке - с конца базы.
    for i := Length(Recs) downto 0 do
    begin
      if (((Recs[i].Sender = AbonentUID) and (Recs[i].recipient = ContactUID)) or
        ((Recs[i].Sender = ContactUID) and (Recs[i].recipient = AbonentUID))) then
      begin
        // Нашли запись, соответствующую абоненту и его собеседнику (один из них отправитель, другой - получатель). Добавляем её в историю.
        Inc(j);
        SetLength(History, j);
        History[j - 1] := Recs[i];
        // Отмечаем выдачу сообщения получателю, если этого не было сделано ранее.
        Recs[i].gotByRecipient := True;
        // Ограничиваем выдачу истории десятью сообщениями, чтобы не затормаживать процесс перебора базы.
        if j = 10 then
        begin
          Break;
        end;
      end;
    end;
  end;
end;

procedure TMessagesDB.getNew(abonentUID: string; var Messages: TMessageRecArray);
var
  i, j: longint;
begin
  // Очищаем массив результата.
  SetLength(Messages, 0);
  if Length(Recs) > 0 then
  begin
    // База сообщений не пуста - можно перебирать.
    j := 0;
    // Перебор записей ведём в обратном порядке
    for i := Length(Recs) downto 0 do
    begin
      if (Recs[i].recipient = AbonentUID) then
      begin
        // в цикле идём до последнего полученного ранее сообщения; его в новые уже не выдаём
        if Recs[i].gotByRecipient then
        begin
        {
        // на всякий случай проверяем, не затесалось ли предыдущее сообщение
        if (i > 0) then
         if not Recs[i - 1].gotByRecipient then
          begin
           SetLength(Messages, j);
           Messages[j - 1] := Recs[i - 1];
          end;
         }
          Break;
        end;
        // Добавляем сообщение в результат
        Inc(j);
        SetLength(Messages, j);
        Messages[j - 1] := Recs[i];
        // Отмечаем "прочитанность" сообщения, если этого не было сделано ранее
        Recs[i].gotByRecipient := True;
      end;
    end;
  end;
end;

function TMessagesDB.indexByUID(UID: string): integer;
var
  i: integer;
begin
  // Устанавливаем результат по умолчанию в -1
  Result := -1;
  if Length(Recs) > 0 then
  begin
    // История не пуста - можно перебирать.
    i := -1;
    // Перебор делаем с начала циклом repeat
    repeat
      Inc(i);
      if (Recs[i].UID = UID) then
      begin
        Result := i;
      end;
    until (Result <> -1) or (i = Length(Recs) - 1);
  end;
end;

procedure TMessagesDB.addMessage(when: int64; senderUID, recipientUID: string;
  msgType: byte; msgText: string; var recIndex: integer; var recUID: string);
begin
  // Генерируем уникальный идентификатор нового сообщения.
  recUID := generateID;
  // Сразу получаем и "столбим" под него место в массиве.
  recIndex := Length(Recs);
  SetLength(Recs, recIndex + 1);
  // Заполняем свойства. Последним ставим отправителя и получателя, чтобы избежать досрочной выдачи сообщения в новые и историю.
  Recs[recIndex].uid := recUID;
  Recs[recIndex].Text := msgText;
  Recs[recIndex].msgType := msgType;
  Recs[recIndex].when := when;
  Recs[recIndex].gotByRecipient := False;
  Recs[recIndex].Sender := senderUID;
  Recs[recIndex].recipient := recipientUID;
end;

procedure TMessagesDB.readFromFile;
var
  f: TextFile;
  n: string;
  i: integer;
begin
  Init;
  n := ExtractFilePath(ParamStr(0)) + 'database/messages';
  if FileExists(n) then
  begin
    AssignFile(f, n);
    reset(f);
    while not EOF(f) do
    begin
      i := Length(Recs);
      SetLength(Recs, i + 1);
      readln(f, Recs[i].uid);
      readln(f, Recs[i].when);
      readln(f, Recs[i].Sender);
      readln(f, Recs[i].recipient);
      readln(f, n);
      if n = '1' then
        Recs[i].gotByRecipient := True
      else
        Recs[i].gotByRecipient := False;
      readln(f, Recs[i].msgType);
      readln(f, Recs[i].Text);
    end;
    CloseFile(f);
  end;
end;

procedure TMessagesDB.saveToFile;
var
  f: TextFile;
  i: integer;
begin
  if Length(Recs) > 0 then
  begin
    AssignFile(f, ExtractFilePath(ParamStr(0)) + 'database/messages');
    rewrite(f);
    for i := 0 to Length(Recs) - 1 do
    begin
      writeln(f, Recs[i].uid);
      writeln(f, Recs[i].when);
      writeln(f, Recs[i].Sender);
      writeln(f, Recs[i].recipient);
      if Recs[i].gotByRecipient then
        writeln(f, '1')
      else
        writeln(f, '0');
      writeln(f, Recs[i].msgType);
      writeln(f, Recs[i].Text);
    end;
    CloseFile(f);
  end;
end;

end.
