unit db_sessions;

{
Программа "Видеочат"
Copyright (c) 2020 Любезный В. В.
https://www.lubezniy.ru/soft/videochat

Модуль описания класса базы данных пользовательских сессий.
}

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils;

type

  // Тип записи сессии
  TSessionRec = record
    SessionUID: string[32];
    // Уникальный идентификатор пользовательской сессии
    UserUID: string[32];
    // Уникальный идентификатор пользователя, к которому относится сессия.
    Active: boolean;
    // статус сессии (False - недействующая, True - действующая)
  end;

  // Тип объекта базы данных сессии
  TSessionsDB = object

    // Массив записей сессий
    Recs: array of TSessionRec;

    // Метод инициализации БД
    procedure Init;

    // Метод чтения БД из файла
    procedure readFromFile;

    // Метод записи БД в файл
    procedure writeToFile;

    {
     Метод создания новой сессии.

     Параметры:

      UserUID - уникальный идентификатор пользователя, для которого мы добавляем сессию.

     Возвращается уникальный идентификатор сессии.
    }
    function newSession(UserUID: string): string;
    // добавление сессии пользователю; возвращается id новой сессии

    {
     Метод определения пользователя по уникальному идентификатору сессии.

     Параметры:

      SessionUID - уникальный идентификатор сессии.

     Возвращается уникальный идентификатор пользователя. Если сессия с таким идентификатором в базе отсутствует, возвращается пустая строка.
    }
    function getUserUID(SessionUID: string): string;

    {
     Метод удаления сессии.

     Параметры:

      SessionUID - уникальный идентификатор сессии.

    }
    procedure deleteSession(SessionUID: string);
  end;

implementation

uses stringfuncs;

procedure TSessionsDB.Init;
begin
  SetLength(Recs, 0);
end;

procedure TSessionsDB.readFromFile;
var
  f: file of TSessionRec;
  n: string;
  i: integer;
begin
  Init;
  n := ExtractFilePath(ParamStr(0)) + 'database/sessions';
  if FileExists(n) then
  begin
    AssignFile(f, n);
    reset(f);
    while not EOF(f) do
    begin
      i := Length(Recs);
      SetLength(Recs, i + 1);
      Read(f, Recs[i]);
      if not Recs[i].Active then
      begin
        // Неактивные записи сессий не читаем
        SetLength(Recs, i);
      end;
    end;
    CloseFile(f);
  end;
end;

procedure TSessionsDB.writeToFile;
var
  f: file of TSessionRec;
begin
  if Length(Recs) > 0 then
  begin
    AssignFile(f, ExtractFilePath(ParamStr(0)) + 'database/sessions');
    rewrite(f);
    BlockWrite(f, Recs[0], Length(Recs));
    CloseFile(f);
  end;
end;

function TSessionsDB.newSession(UserUID: string): string;
var
  i: integer;
begin
  i := Length(Recs);
  SetLength(Recs, i + 1);
  // Генерируем уникальный идентификатор сессии и подставляем его в выдачу.
  Result := generateID;
  // Заполняем запись сессии
  Recs[i].UserUID := UserUID;
  Recs[i].SessionUID := Result;
  // Сессию устанавливаем сразу как активную
  Recs[i].Active := True;
  writeToFile;
end;

function TSessionsDB.getUserUID(SessionUID: string): string;
var
  i: integer;
begin
  Result := '';
  if (Length(Recs) > 0) then
  begin
    for i := 0 to Length(Recs) - 1 do
    begin
      if ((Recs[i].SessionUID = SessionUID) and (Recs[i].Active)) then
      begin
        Result := Recs[i].UserUID;
        Break;
      end;
    end;
  end;
end;

procedure TSessionsDB.deleteSession(SessionUID: string);
var
  i: integer;
begin
  if (Length(Recs) > 0) then
  begin
    // ищем запись с нужным идентификатором
    for i := 0 to Length(Recs) - 1 do
    begin
      if (Recs[i].SessionUID = SessionUID) then
      begin
        // Устанавливаем найденную сессию как неактивную
        Recs[i].Active := False;
        Break;
      end;
    end;
  end;
  // сохраняем результат в файл
  writeToFile;
end;

end.
