unit uhttpsserver;

{
Программа "Видеочат"
Copyright (c) 2020 Любезный В. В.
https://www.lubezniy.ru/soft/videochat

Модуль описания классов Web-сервера и обработчика HTTPS-запросов.
Разработан на базе примера HTTPS-сервера из пакета Synapse для Lazarus.
}

{$mode objfpc}{$H+}

interface

uses
  Classes, blcksock, synsock, Synautil, SysUtils, Dialogs, webserverconfig_class,
  requestheaders_class, postdata_class;

type

  {
  Описание класса Web-сервера
  }
  TTCPHttpDaemon = class(TThread)
  private
    { Объект конфигурации }
    Config: TWebServerConfig;
    { Объект сокета }
    Sock: TTCPBlockSocket;
    { Старая переменная окружения OPENSSL_CONF }
    OldOpenSSLCfgEnvVar: UnicodeString;
  public
    constructor Create(Cfg: TWebServerConfig);
    destructor Destroy; override;
    { Запуск Web-сервера }
    procedure Execute; override;
  end;

  {
  Описание класса обрабатываемого Web-сервером запроса (в многопоточном режиме)
  }
  TTCPHttpThrd = class(TThread)
  private
    { Частные свойства }
    { Объект конфигурации }
    Config: TWebServerConfig;
    { Объект сокета }
    Sock: TTCPBlockSocket; // объект сокета
    { URI запроса }
    URI: string;
    { Строка GET-параметров запроса (после ? в URL) }
    QueryString: string;
    { GET-параметры запроса (application/x-www-form-urlencoded, multipart/form-data, application/json) }
    GetParams: array of TPOSTDataRec;
    { POST-параметры и данные запроса }
    POSTData: TPOSTData;
    { Частные методы }
    { Выдача HTTP-ответа с кодом 400 (Bad Request) }
    procedure BadRequest;
    { Парсинг GET-параметров запроса }
    procedure parseGETParams(QueryLine: string);
    // парсинг параметров запроса
  public
    { Публичные свойства }
    RequestHeaders: TRequestHeaders;
    { Объект списка HTTP-заголовков запроса и ответа }
    Headers: TStringList;
    { Стримы входящих и выходящих данных }
    InputData, OutputData: TMemoryStream;
    { Публичные методы }
    constructor Create(HSock: TSocket; Cfg: TWebServerConfig);
    destructor Destroy; override;
    { Метод парсинга и запуска обработки HTTPS-запроса }
    procedure Execute; override;
    { Метод запроса GET-параметра http-запроса }
    function QueryParamValue(Name: string): string;
    { Метод запроса POST-параметра http-запроса }
    function POSTParamValue(Name: string): string;
    { Основной метод web-приложения - обработка HTTP-запроса; возвращается http-код }
    function ProcessHttpRequest(Request, QueryLine: string): integer;
  end;

implementation

uses stringfuncs, chatapptypes, fpjson, ssl_openssl, Windows;

var
  {
   Флаг отладочного режима. Значение берётся из параметра DebugMode в config/webserver.json.
   При значении true по ходу работы в папке с программой создаётся файл debug.txt, куда пишется информация.
   Пригоден только при отладке одиночных запросов, иначе будут конфликты при попытке одновременно записать файл с соответствующими ошибками.
  }
  DebugMode: boolean;

{
Процедура записи сообщения в отладочный лог
}
procedure writeFileMsg(Msg: string);
var
  fn: string;
  f: TextFile;
begin
  if DebugMode then
  begin
    fn := ExtractFilePath(ParamStr(0)) + '/debug.txt';
    AssignFile(f, fn);
    if FileExists(fn) then
      Append(f)
    else
      rewrite(f);
    writeln(f, Msg);
    CloseFile(f);
  end;
end;

{ TTCPHttpDaemon - класс Web-сервера }

{
Конструктор класса
}
constructor TTCPHttpDaemon.Create(Cfg: TWebServerConfig);
var
  s: string;
begin
  // Запоминаем у себя объект конфигурации.
  Config := Cfg;
  // Устанавливаем флаг отладочного режима. Значение берём из объекта конфигурации.
  DebugMode := Config.DebugMode;
  // Важно! Устанавливаем в переменной окружения путь к файлу конфигурации OpenSSL.
  OldOpenSSLCfgEnvVar := SysUtils.GetEnvironmentVariable('OPENSSL_CONF');
  if OldOpenSSLCfgEnvVar = '' then
  begin
    s := Cfg.OpenSSLConfigFile;
    SetEnvironmentVariable('OPENSSL_CONF', PChar(s));
  end;
  // Создаём TCP-сокет.
  Sock := TTCPBlockSocket.Create;
  // Помним, что при завершении сервера нам нужно освободить память
  FreeOnTerminate := True;
  inherited Create(False);
end;

{
Деструктор класса
}
destructor TTCPHttpDaemon.Destroy;
begin
  // освобождаем объект сокета
  FreeAndNil(Sock);
  inherited Destroy;
end;

{
Метод запуска Web-сервера
}
procedure TTCPHttpDaemon.Execute;
begin
  with Sock do
  begin
    // создаём серверный сокет
    CreateSocket;
    // задаём поведение сокета и таймаут
    setLinger(Config.LingerEnable, Config.LingerTimeout);
    // связываем сокет с заданными IP-адресом интерфейса сервера (0.0.0.0 - все адреса) и портом (443 - протокол HTTPS)
    bind(Config.BindAddress, IntToStr(Config.BindPort));
    // запускаем слушание порта и ждём данных
    Listen;
    // основной рабочий цикл
    repeat
      // завершение работы сервера - завершаем цикл
      if Terminated then
        break;
      // ожидание данных (таймаут задаётся в миллисекундах)
      if CanRead(Config.CanReadTimeout) then
      begin
        // если ошибки нет, создаём объект входящего подключения, сразу передавая сокет, с которым соединён клиент
        if LastError = 0 then
        begin
          TTCPHttpThrd.Create(Accept, Config);
          { вместо Accept было ClientSock  }
        end
        else
        begin
          writeFileMsg('Error after canRead. lastError ' + IntToStr(LastError));
        end;
      end;
      // конец основного рабочего цикла
    until False;
  end;
end;

{ TTCPHttpThrd - объект входящего запроса (thread)}

{
Конструктор
Параметры:
HSock - сокет подключения
}
constructor TTCPHttpThrd.Create(HSock: TSocket; Cfg: TWebServerConfig);
begin
  // создаём объекты
  SetLength(GetParams, 0);
  // объект сокета
  Sock := TTCPBlockSocket.Create;
  // объект заголовков запроса
  RequestHeaders := TRequestHeaders.Create;
  // объект POST-данных
  POSTData := TPOSTData.Create;
  // строки - заголовок запроса
  Headers := TStringList.Create;
  // стрим для входящих данных
  InputData := TMemoryStream.Create;
  // стрим для исходящих данных
  OutputData := TMemoryStream.Create;
  // цепляемся к сокету
  Sock.Socket := HSock;
  // говорим, что при завершении обработки надо освободить память
  FreeOnTerminate := True;
  // запоминаем у себя объект конфигурации
  Config := Cfg;
  // выполняем всё остальное
  inherited Create(False);
end;

{
Деструктор
}
destructor TTCPHttpThrd.Destroy;
begin
  // освобождаем всё, что мы насоздавали в Create
  FreeAndNil(Sock);
  FreeAndNil(Headers);
  FreeAndNil(InputData);
  FreeAndNil(OutputData);
  FreeAndNil(RequestHeaders);
  FreeAndNil(POSTData);
  SetLength(GetParams, 0);
  // отрабатываем остальное
  inherited Destroy;
end;

{
Возврат ответа с кодом 400
}
procedure TTCPHttpThrd.BadRequest;
begin
  Sock.SendString('HTTP/1.0 400 Bad Request' + CRLF + CRLF +
    '<html><h1>Bad Request</h1></html>');
end;

{
Метод обработки входящего соединения. Переопределён от родительского класса
}
procedure TTCPHttpThrd.Execute;
var
  s: string;
  Method, URIWithParams, Protocol: string;
  Size: integer;
  x, n: integer;
  ResultCode: integer;
begin
  writeFileMsg('Thread execute started');

  // Установка параметров серверного сертификата SSL.

  // путь к файлу корневого сертификата удостоверяющего центра
  if (Pos('/', Config.CACertificateFile) = 0) and
    (Pos('\', Config.CACertificateFile) = 0) then
    Sock.SSL.CertCAFile := ExtractFilePath(ParamStr(0)) + Config.CACertificateFile
  else if Config.CACertificateFile <> '' then
    Sock.SSL.CertCAFile := Config.CACertificateFile;
  // путь к файлу сертификата открытого ключа, подписанного удостоверяющим центром
  if (Pos('/', Config.CertificateFile) = 0) and
    (Pos('\', Config.CertificateFile) = 0) then
    Sock.SSL.CertificateFile := ExtractFilePath(ParamStr(0)) + Config.CertificateFile
  else if Config.CertificateFile <> '' then
    Sock.SSL.CertificateFile := Config.CertificateFile;
  // путь к файлу закрытого ключа
  if (Pos('/', Config.PrivateKeyFile) = 0) and (Pos('\', Config.PrivateKeyFile) = 0) then
    Sock.SSL.PrivateKeyFile := ExtractFilePath(ParamStr(0)) + Config.PrivateKeyFile
  else if Config.PrivateKeyFile <> '' then
    Sock.SSL.PrivateKeyFile := Config.PrivateKeyFile;
  // пароль на закрытый ключ
  if Config.PrivateKeyPassword <> '' then
    Sock.SSL.KeyPassword := Config.PrivateKeyPassword;

  writeFileMsg('CA cert file ' + Sock.SSL.CertCAFile);
  writeFileMsg('Cert file ' + Sock.SSL.CertificateFile);
  writeFileMsg('Key file ' + Sock.SSL.PrivateKeyFile);
  writeFileMsg('Password ' + Sock.SSL.KeyPassword);

  // флаг проверки SSL-сертификата клиента
  Sock.SSL.VerifyCert := Config.VerifySSL;

  // принимаем соединение, если получается
  try
    if (not Sock.SSLAcceptConnection) or (Sock.SSL.LastError <> 0) then
    begin
      if not Sock.SSLAcceptConnection then
      begin
        s := 'Sock.SSLAcceptConnection = False';
      end
      else
      begin
        s := 'Sock.SSLAcceptConnection = True';
      end;
      s := s + ' Sock.SSL.LastError ' + IntToStr(Sock.SSL.LastError) +
        ' ' + Sock.SSL.LastErrorDesc;
      writeFileMsg('SSL error: ' + s);
      Exit;
    end;
  except
    on E: Exception do
    begin
      writeFileMsg('SSL exception: ' + E.Message);
      Exit;
    end;
  end;
  // чтение строки запроса (первая строка - например, GET /index.html HTTP/1.1) в переменную s
  s := Sock.RecvString(Config.ReceiveTimeout);
  // при ошибке завершаем поток
  if Sock.LastError <> 0 then
    Exit;
  // при пустых данных тоже
  if s = '' then
    Exit;
  // парсим строку; сначала получаем метод запроса (параметр обязательный)
  Method := Fetch(s, ' ');
  if (s = '') or (Method = '') then
  begin
    BadRequest;
    Exit;
  end;
  // теперь URI (параметр также обязательный)
  URIWithParams := Fetch(s, ' ');
  if URIWithParams = '' then
  begin
    BadRequest;
    Exit;
  end;
  // теперь протокол
  Protocol := Fetch(s, ' ');
  // если протокол пуст или не http, посылаем
  if Protocol = '' then
    Exit;
  if pos('HTTP/', Protocol) <> 1 then
    Exit;
  // начинаем сбор http-заголовков в TStringList
  Headers.Clear;
  Size := -1;
  // чтение заголовков запроса
  repeat
    // читаем строку
    s := Sock.RecvString(Config.ReceiveTimeout);
    // при ошибке выходим
    if Sock.LastError <> 0 then
      Exit;
    if s <> '' then
    begin
      Headers.Add(s);
      // сразу отслеживаем Content-Length
      if Pos('CONTENT-LENGTH:', Uppercase(s)) = 1 then
        Size := StrToIntDef(SeparateRight(s, ' '), -1);
    end;
  until s = '';
  // при размере больше максимально допустимого сразу посылаем
  if (Size > Config.MaxRequestBodySize) then
  begin
    BadRequest;
    Exit;
  end;
  // заполняем объект заголовков запроса
  RequestHeaders.Read(Headers);
  // принимаем тело запроса, если оно есть
  InputData.Clear;
  if Size >= 0 then
  begin
    InputData.SetSize(Size);
    x := Sock.RecvBufferEx(InputData.Memory, Size, Config.ReceiveTimeout);
    InputData.SetSize(x);
    if Sock.LastError <> 0 then
    begin
      Exit;
    end;
    if x > Config.MaxRequestBodySize then
    begin
      BadRequest;
      Exit;
    end;
  end;
  // очищаем выходные данные
  OutputData.Clear;
  // обрабатываем HTTP-запрос и получаем HTTP-код ответа
  ResultCode := ProcessHttpRequest(Method, URIWithParams);
  writeFileMsg('Result code ' + IntToStr(ResultCode));
  if Protocol <> '' then
  begin
    // отправляем первую строку ответа
    Sock.SendString('HTTP/1.0 ' + IntToStr(ResultCode) + CRLF);
    if Sock.LastError <> 0 then
    begin
      Exit;
    end;
    // добавляем заголовки ответа
    Headers.Add('Content-Length: ' + IntToStr(OutputData.Size));
    Headers.Add('Connection: close');
    Headers.Add('Date: ' + Rfc822DateTime(now));
    Headers.Add('Server: Synapse HTTP server demo modified');
    Headers.Add('');
    // отправляем заголовки ответа
    for n := 0 to Headers.Count - 1 do
    begin
      Sock.SendString(Headers.Strings[n] + CRLF);
      if Sock.LastError <> 0 then
      begin
        Exit;
      end;
    end;
  end;
  // отправляем ответ
  Sock.SendBuffer(OutputData.Memory, OutputData.Size);
  writeFileMsg('Thread execute ended');
end;

function TTCPHttpThrd.QueryParamValue(Name: string): string;
var
  i: integer;
begin
  Result := '';
  if Length(GetParams) > 0 then
    for i := 0 to Length(GetParams) - 1 do
      if (GetParams[i].Name = Name) then
      begin
        Result := GetParams[i].Value;
        Break;
      end;
end;

function TTCPHttpThrd.POSTParamValue(Name: string): string;
var
  i: integer;
begin
  Result := '';
  if POSTData <> nil then
    if Length(POSTData.Recs) > 0 then
      for i := 0 to Length(POSTData.Recs) - 1 do
        if (POSTData.Recs[i].Name = Name) then
        begin
          Result := POSTData.Recs[i].Value;
          Break;
        end;
end;

procedure TTCPHttpThrd.parseGETParams(QueryLine: string);
var
  p: integer;
  s, t: string;
  Param: TPOSTDataRec;
begin
  URI := QueryLine;
  QueryString := '';
  p := Pos('?', QueryLine);
  if p > 0 then
  begin
    URI := Copy(QueryLine, 1, p - 1);
    QueryString := Copy(QueryLine, p + 1, Length(QueryLine) - p);
  end;
  if QueryString <> '' then
  begin
    s := QueryString;
    // Выделяем и обрабатываем параметры в цикле.
    while s <> '' do
    begin
      // В переменную t кладём параметр и отделяем его от следующего (по позиции знака &)
      p := Pos('&', s);
      if p <> 0 then
      begin
        t := Copy(s, 1, p - 1);
        s := Copy(s, p + 1, Length(s) - p);
      end
      else
      begin
        t := s;
        s := '';
      end;
      // Отделяем по знаку = имя параметра от его значения.
      p := Pos('=', t);
      if p <> 0 then
      begin
        // До = у нас имя, а после - значение (URL-encoded)
        Param.Name := Copy(t, 1, p - 1);
        Param.Value := URLDecode(Copy(t, p + 1, Length(t) - p));
      end
      else
      begin
        // Знака = нет - параметр с пустым значением.
        Param.Name := t;
        Param.Value := '';
      end;
      // Добавляем параметр в список
      p := Length(GetParams);
      SetLength(GetParams, p + 1);
      GetParams[p] := Param;
    end;
  end;
end;

function TTCPHttpThrd.ProcessHttpRequest(Request, QueryLine: string): integer;
var
  // вспомогательные переменные
  l: TStringList;
  jo: TJSONObject;
  i, p: integer;
  s: string;
  ts: int64;
  b: boolean;

  // основные переменные
  ContentType: string; // HTTP-заголовок Content-Type
  currentUserIndex: integer;
  // индекс записи текущего пользователя в базе
  currentSession: string;
  // текущее значение сессионной cookie, если таковая есть
  contactUID: string;
  // строка нашего собеседника (для получения истории)
  currentTimestamp: int64; // UNIX timestamp текущего времени
  chatReqType: string; // тип запроса
  msgType: integer; // тип сообщения
  msgData: string; // текст сообщения
begin
  {
  Обработка http-запроса.
  Request - метод запроса (GET, POST и т. д.)
  QueryLine - строка запроса (путь и параметры)
  InputData (TMemoryStream) - тело запроса.
  RequestHeaders (TRequestHeaders) - заголовки запроса
  OutputData (TMemoryStream) - контент ответа. Заполняем с помощью L: TStringList
  Result - HTTP-код ответа.
  С помощью метода QueryParamValue можно получить GET-параметр запроса по имени.
  С помощью метода POSTParamValue можно получить POST-параметр запроса по имени.
  С помощью RequestHeaders.getValue можно получить значение параметра запроса.
  С помощью RequestHeaders.getCookieValue можно получить значение куки.
  }
  writeFileMsg('processHTTPRequest is started');
  // парсим GET-параметры запроса, если есть
  parseGETParams(QueryLine);
  writeFileMsg('parseGETParams is completed');
  // получаем текущий timestamp
  currentTimestamp := Trunc((Now - EncodeDate(1970, 1, 1)) * 24 * 60 * 60);
  writeFileMsg('got currentTimestamp');
  // чистим заголовки ответа
  Headers.Clear;
  // устанавливаем дефолтный код http-ответа
  Result := 504;
  writeFileMsg('Checking authentication');
  // проверяем, аутентифицирован ли пользователь, по сессии, и заполняем переменную CurrentUserIndex
  if not RequestHeaders.getCookieValue('session', currentSession) then
  begin
    currentUserIndex := -1;
  end
  else
  begin
    currentUserIndex := ChatApp.getUserIndexBySession(currentSession);
    if currentUserIndex <> -1 then
      ChatApp.UserDB.Recs[currentUserIndex].lastLogin := currentTimestamp;
    currentSession := '';
  end;

  // Делаем обработку в зависимости от методов
  if (Request = 'GET') then
  begin
    // Обработка запроса методом GET
    writeFileMsg('processing GET request');
   {
   Задача: При обработке корня неавторизованному пользователю выдать страницу входа в систему, авторизованному - страницу чата.
   При обработке запроса /users отдать json со списком пользователей. Туда отдать uid, display name и статус (online, если логинился в последние 10 сек, или offline)
   По остальным запросам обработать статику; в случае ненайденного выдать 404.
   }
    if (URI = '/') then
    begin
      // обработка запроса корня (/)
      writeFileMsg('GET / started');
      if currentUserIndex = -1 then
      begin
        // пользователь не авторизован - выдаём страницу логина
        i := Config.getStaticFileIndex('/login', False);
        Headers.Add('Content-Type: ' + Config.StaticFileURIs[i].ContentType);
        OutputData.LoadFromFile(ExtractFilePath(ParamStr(0)) +
          Config.StaticFileURIs[i].FilePath);
        Result := 200;
      end
      else
      begin
        // пользователь авторизован - выдаём страницу чата
        i := Config.getStaticFileIndex('/main', True);
        Headers.Add('Content-Type: ' + Config.StaticFileURIs[i].ContentType);
        OutputData.LoadFromFile(ExtractFilePath(ParamStr(0)) +
          Config.StaticFileURIs[i].FilePath);
        Result := 200;
      end;
      writeFileMsg('GET / ended');
      // конец обработки запроса корня
    end
    else if (URI = '/contacts') then
    begin
      // Обработка запроса /contacts - получение списка контактов.
      writeFileMsg('GET /contacts started');
      l := TStringList.Create;
      Headers.Add('Content-Type: application/json');
      l.Clear;
      // Неавторизованным пользователям выдаём сообщение об ошибке.
      if currentUserIndex = -1 then
      begin
        l.Add('{"code": 1, "message": "Необходима авторизация."}');
      end
      else
      begin
        // Для авторизованных собираем список контактов.
        ts := currentTimestamp - 10;
        b := False;
        l.Add('{"code": 0, "current": {"UID": "' +
          ChatApp.UserDB.Recs[currentUserIndex].UID + '", "displayName": "' +
          str_replace(ChatApp.UserDB.Recs[currentUserIndex].DisplayName, '"', '\"', p) +
          '"}, "data": [');
        for i := 0 to Length(ChatApp.UserDB.Recs) - 1 do
          if i <> currentUserIndex then
          begin
            // Нашли контакт - добавляем
            // Сначала запятую-разделитель, если это не первая запись
            s := '';
            if b then
              s := ','
            else
              b := True;
            // Теперь формируем JSON-объект записи, заполняем его данными, формируем и добавляем JSON-строку к s
            jo := TJSONObject.Create;
            jo.add('uid', ChatApp.UserDB.Recs[i].UID);
            jo.add('displayName', ChatApp.UserDB.Recs[i].DisplayName);
            jo.add('online', (ChatApp.UserDB.Recs[i].lastLogin >= ts));
            // Заполнили - добавляем JSON-строку к s
            s := s + jo.AsJSON;
            // Освобождаем память, уничтожая объект JSON
            FreeAndNil(jo);
            // Добавляем s в результирующий TStringList
            l.Add(s);
          end;
        l.Add(']}');
        writeFileMsg('GET /contacts ended');
      end;
      l.SaveToStream(OutputData);
      l.Free;
      Result := 200;
      // конец обработки запроса /contacts
    end
    else if (URI = '/history') then
    begin
      // запрос /history - получить историю чата с заданным абонентом
      // получаем из запроса UID нашего собеседника
      contactUID := queryParamValue('contact');
      // если такового нет, не пущаем
      if contactUID = '' then
        Result := 404
      else
      if currentUserIndex = -1 then
      begin
        // пользователь не авторизован
        l := TStringList.Create;
        l.Add('{"code": 1, "message": "Необходима авторизация."}');
        Headers.Add('Content-Type: application/json');
        l.SaveToStream(OutputData);
        l.Free;
        Result := 200;
      end
      else
      begin
        // получаем историю
        l := ChatApp.getChatHistory(ChatApp.UserDB.Recs[currentUserIndex].UID,
          contactUID, currentTimestamp);
        Headers.Add('Content-Type: application/json');
        l.SaveToStream(OutputData);
        l.Free;
        Result := 200;
      end;
      // конец обработки запроса history
    end
    else if (URI = '/newmessages') then
    begin
      // обработка запроса /newmessages - получение новых сообщений
      if currentUserIndex = -1 then
      begin
        // пользователь не авторизован
        Headers.Add('Content-Type: application/json');
        l := TStringList.Create;
        l.Add('{"code": 1, "message": "Необходима авторизация."}');
        Headers.Add('Content-Type: application/json');
        l.SaveToStream(OutputData);
        l.Free;
        Result := 200;
      end
      else
      begin
        // получаем данные
        l := ChatApp.getNewMessages(ChatApp.UserDB.Recs[currentUserIndex].UID);
        Headers.Add('Content-Type: application/json');
        l.SaveToStream(OutputData);
        l.Free;
        Result := 200;
      end;
      // конец обработки запроса newmessages
    end
    else if (URI = '/logout') then
    begin
      // запрос logout - выход из системы
      // Проверяем авторизацию.
      if currentUserIndex = -1 then
      begin
        // Пользователь не авторизован. Ругаемся не очень понятными вещами.
        Headers.Add('Content-Type: application/json');
        l := TStringList.Create;
        l.Add('{"code": 1, "message": "Необходима авторизация."}');
        Headers.Add('Content-Type: application/json');
        l.SaveToStream(OutputData);
        l.Free;
        Result := 200;
      end
      else
      begin
        // Пользователь авторизован. Удаляем сессию, очищаем сессионный cookie и перенаправляем пользователя на главную.
        ChatApp.SessionsDB.deleteSession(currentSession);
        // Очистка cookie выполняется путём её установки с параметром Expires заведомо меньше текущего времени.
        Headers.Add('Set-Cookie: session=aaa; Secure; HttpOnly; Expires=' +
          RFC1123TimeFormat(Now - 30, True));
        // Редирект
        l := TStringList.Create;
        l.Add('<html>');
        l.Add('<head>');
        l.Add('<script type="text/javascript">');
        l.Add('function start() ' + Chr(123) + ' document.location.href = "/"; ' +
          Chr(125));
        l.Add('</script>');
        l.Add('</head>');
        l.Add('<body onLoad="start()"></body>');
        l.Add('</html>');
        l.SaveToStream(OutputData);
        l.Free;
        Result := 200;
      end;
      // Конец обработки запроса logout
    end
    else
    begin
      // URL отличается от ранее заданных - включаем обработчик статики
      i := Config.getStaticFileIndex(URI, (CurrentUserIndex <> -1));
      if i = -1 then
      begin
        Result := 404;
      end
      else
      begin
        Headers.Add('Content-Type: ' + Config.StaticFileURIs[i].ContentType);
        OutputData.LoadFromFile(ExtractFilePath(ParamStr(0)) +
          Config.StaticFileURIs[i].FilePath);
        Result := 200;
      end;
    end;
    // Конец обработчика GET-запросов
  end
  else if (Request = 'POST') then
  begin
    // Обработчик запросов POST
    // Получаем заголовок Content-Type. При отсутствии возвращаем HTTP-ошибку 400 (Bad Request).
    if not RequestHeaders.getValue('Content-Type', ContentType) then
    begin
      Result := 400;
      Exit;
    end;
    // Читаем POST-данные. В случае неудачи возвращаем HTTP-ошибку 400 (Bad Request).
    if not POSTData.Read(InputData, ContentType) then
    begin
      Result := 400;
      Exit;
    end;
    // В зависимости от URI выполняем обработку
    if (URI = '/login') then
    begin
      // URI /login - аутентификация пользователя по логину и паролю (POST-параметры login и password)
      // Получаем индекс записи пользователя в нашей БД.
      CurrentUserIndex := ChatApp.UserDB.Authenticate(POSTParamValue('login'),
        POSTParamValue('password'));
      if CurrentUserIndex = -1 then
      begin
        // С логином-паролем что-то не так, редиректим пользователя обратно (на главную, оттуда отправят на логин).
        l := TStringList.Create;
        l.Add('<html>');
        l.Add('<head>');
        l.Add('<script type="text/javascript">');
        l.Add('function start() ' + Chr(123) + ' document.location.href = "/"; ' +
          Chr(125));
        l.Add('</script>');
        l.Add('</head>');
        l.Add('<body onLoad="start()"></body>');
        l.Add('</html>');
        l.SaveToStream(OutputData);
        l.Free;
        Result := 200;
      end
      else
      begin
        // Делаем пользователю сессию и ставим куку, затем html-ом редиректим на главную.
        s := ChatApp.SessionsDB.newSession(ChatApp.UserDB.Recs[CurrentUserIndex].UID);
        // Куку ставим со сроком действия в 30 дней
        Headers.Add('Set-Cookie: session=' + s + '; Secure; HttpOnly; Expires=' +
          RFC1123TimeFormat(Now + 30, True));
        // Делаем JavaScript-редирект
        l := TStringList.Create;
        l.Add('<html>');
        l.Add('<head>');
        l.Add('<script type="text/javascript">');
        l.Add('function start() ' + Chr(123) + ' document.location.href = "/"; ' +
          Chr(125));
        l.Add('</script>');
        l.Add('</head>');
        l.Add('<body onLoad="start()"></body>');
        l.Add('</html>');
        l.SaveToStream(OutputData);
        l.Free;
        Result := 200;
      end;
      // Конец обработки URI /login
    end
    else if (URI = '/chat') then
    begin
      // URI /chat - отправка сообщений по чату в синхронном и асинхронном режимах
      if currentUserIndex = -1 then
      begin
        // Пользователь не авторизован - ругаемся.
        Headers.Add('Content-Type: application/json');
        l := TStringList.Create;
        l.Clear;
        l.Add('{"code": 1, "message": "Необходима авторизация."}');
        l.saveToStream(OutputData);
        Result := 200;
      end
      else
      begin
        // Пользователь авторизован.
        Result := 200;
        // Здесь у нас данные POST.
        // Получаем тип запроса
        chatReqType := POSTParamValue('type');
        // В зависимости от типа запроса выполняем обработку.
        if chatReqType = 'asynctransmit' then
        begin
          // Обработка запроса асинхронной передачи сообщения; завершаем её, когда запишем сообщение в БД
          // Получаем и валидируем параметры.
          // UID получателя
          i := ChatApp.UserDB.IndexByUID(POSTParamValue('target'));
          if i = -1 then
          begin
            // Контакта не нашли - возвращаем 404
            Result := 404;
            Exit;
          end;
          // Тип сообщения
          msgType := StrToInt(POSTParamValue('msgType'));
          if (msgType < 0) or (msgType > 9) then
          begin
            // Неверный тип сообщения - возвращаем 400 и выходим.
            Result := 400;
            Exit;
          end;
          msgData := POSTParamValue('msgData');
          // Добавляем сообщение в базу.
          ChatApp.MessagesDB.addMessage(currentTimestamp,
            ChatApp.UserDB.Recs[currentUserIndex].UID, POSTParamValue('target'),
            msgType, msgData, i, s);
          // В переменной s у нас uid сообщения, в i - индекс записи.
          headers.Add('Content-Type: application/json');
          l := TStringList.Create();
          l.Clear;
          l.Add('{"code": 0, "timestamp": ' + IntToStr(currentTimestamp) +
            ', "uid": "' + s + '"}');
          l.saveToStream(OutputData);
          l.Free;
          // Обработка завершена успешно.
          Result := 200;
          // конец обработчика асинхронной передачи данных
        end
        else if chatReqType = 'synctransmit' then
        begin
          // Обработка запроса синхронной передачи сообщения; завершается она ответом только после того, как сообщение будет отдано собеседнику.
          // Получаем и валидируем параметры.
          // UID получателя
          i := ChatApp.UserDB.IndexByUID(POSTParamValue('target'));
          if i = -1 then
          begin
            // Контакта не нашли - возвращаем 404
            Result := 404;
            Exit;
          end;
          // Для синхронной передачи обязательно проверяем, чтобы пользователь был онлайн.
          if ChatApp.UserDB.Recs[i].lastLogin < (currentTimestamp - 10) then
          begin
            // Пользователь оффлайн. Возвращаем JSON с ошибкой.
            headers.Add('Content-Type: application/json');
            l := TStringList.Create();
            l.Clear;
            l.Add('{"code": 2, "timestamp": ' + IntToStr(currentTimestamp) +
              ', "message": "Пользователь оффлайн. Синхронная передача данных невозможна."}');
            l.saveToStream(OutputData);
            l.Free;
            // выходим
            Result := 200;
            Exit;
          end;
          // Тип сообщения
          msgType := StrToInt(POSTParamValue('msgType'));
          if (msgType < 0) or (msgType > 9) then
          begin
            // Неверный тип сообщения - возвращаем 400.
            Result := 400;
            Exit;
          end;
          msgData := POSTParamValue('msgData');
          // Добавляем сообщение в базу
          ChatApp.MessagesDB.addMessage(currentTimestamp,
            ChatApp.UserDB.Recs[currentUserIndex].UID, POSTParamValue('target'),
            msgType, msgData, i, s);
          // В переменной s у нас uid сообщения, в i - индекс записи.
          // Ожидаем отдачи сообщения получателю.
          b := False;
          p := 0;
          repeat
            // статус сообщения проверяем ежесекундно
            Sleep(1000);
            // и считаем кол-во проверок
            Inc(p);
            // проверяем, не уехали ли у нас сообщения при очистке БД
            if (i >= Length(ChatApp.MessagesDB.Recs)) then
              i := ChatApp.MessagesDB.indexByUID(s)
            else if (ChatApp.MessagesDB.Recs[i].UID <> s) then
              i := ChatApp.MessagesDB.indexByUID(s);
            // уехали, сообщение удалилось - работать дальше смысла нет
            if (i = -1) then
              b := True
            // увидели, что сообщение доставлено получателю
            else if ChatApp.MessagesDB.Recs[i].gotByRecipient then
              b := True
            // цикл проверки выполняем максимум 100 раз, чтобы не завалить приложение
            else if (p > 100) then
              b := True;
          until b;
          // Дождались отдачи или нарвались на ошибку - проверять не будем, просто вернём успех.
          Headers.Add('Content-Type: application/json');
          l := TStringList.Create();
          l.Clear;
          l.Add('{"code": 0, "timestamp": ' + IntToStr(currentTimestamp) +
            ', "uid": "' + s + '"}');
          l.saveToStream(OutputData);
          l.Free;
          // Обработка завершена.
          Result := 200;
        end
        else
        begin
          // Тип запроса определить не удалось - вернём для приличия 404.
          Result := 404;
          Exit;
        end;
      end;
      // Конец обработки URI /chat
    end
    else
    begin
      // URI неизвестен - возвращаем стандартное HTTP 404
      Result := 404;
    end;
    // Конец обработчика POST-запросов
  end
  else if (Request = 'OPTIONS') then
  begin
    // обработка запроса OPTIONS
    Result := 201;
  end;
end;

end.
